//+------------------------------------------------------------------+
//|                                                     Profiler.mqh |
//|                                 Copyright  2011-2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes:Self-made profiler on static & automatic objects |
//|                        https://www.mql5.com/en/blogs/post/682673 |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Global defines                                                   |
//+------------------------------------------------------------------+

// a shortcut for profiling entire function by automatic object of ProfilerObject class
#define _PROFILEFUNC_  static ProfilerLocator locator1(__FUNCTION__); ProfilerObject obj1(locator1);
// a shortcut for profiling entire MQL program
#define _PROFILESCRIPT_  static ProfilerLocator locator2(); ProfilerObject obj2(locator2);

string ProfilerPrefix = "PFP-";

//+------------------------------------------------------------------+
//| Classes declarations                                             |
//+------------------------------------------------------------------+

// forward declaration
class ProfilerLocator;

// main class
class Profiler
{
  private:
    // maximum number of code fragments for measurement, size of arrays below
    int MaxID;
    // current (latest) free index within profiling arrays
    int CurID;
    // last start and stop times for every code fragment
    ulong Start[], Stop[];
    // number of passes made through every code fragment
    ulong CounterTotal[];
    // number of milliseconds spent in every code fragment
    ulong DurationTotal[];
    // descriptive name for every code fragment
    string Names[];
    // a flag, if code fragment was actually used for profiling
    bool Used[];
    // main instance of Profiler
    static Profiler *pSingleton;
    // time at object constructor
    ulong InstanceCreateTime;
    // a flag to save cumulative statistics to global variables
    bool UseGlobalVariables;

  public:
    // constructor accepts a number of measurement elements for code fragments
    // and allocates all private arrays accordingly
    Profiler(int max = 100);
    ~Profiler();
    
    // GetSingleton provides the first (and usually the only) instance of Profiler which is used
    // by ProfilerLocator by default, if another instance is not specified
    static Profiler *GetSingleton()
    {
      return(pSingleton);
    };
    int GetCurrentID(void)
    {
      return(CurID);
    };
    // set name for code fragment with given id
    // return true on success
    bool SetName(int id, string name);
    // set name for next empty code fragment (CurID)
    // return identifier assigned to new code fragment
    int SetName(string name);
    // get name for specified code fragment by its id
    string GetName(int id) const;
    // start profiling code fragment with given id
    // return true on success
    bool StartProfiler(int id = 0);
    // stop profiling code fragment with given id
    // return time (in ms) spent in code fragment for the last single pass
    ulong StopProfiler(int id = 0);
    // start profiling code fragment with an id, bound to given locator
    // return true on success
    bool StartProfiler(const ProfilerLocator &locator);
    // stop all profilers
    void StopProfiling(void);
    // stop profiling code fragment with an id, bound to given locator
    // return time (in ms) spent in code fragment for the last single pass
    ulong StopProfiler(const ProfilerLocator &locator);
    // [depreceted] start profiling code fragment identified by its descriptive name
    // return identifier assigned to newly registered or old (if exists before)
    // code fragment with given name
    int StartProfiler(string name);
    // get count of passes made through code fragment with given id
    ulong GetProfilerCount(int id = 0) const;
    // get overall duration (in ms) of execution of code fragment with given id
    ulong GetProfilerDuration(int id = 0) const;
    // get string description of statistics gathered for code fragment with given id
    // in format "name count seconds (milliseconds)"
    string GetStats(int id = 0) const;
    // print statistics for all used code fragments in the log
    void PrintAllStats(void) const;
    void SetGlobalVariablesUse(bool use);
};

static Profiler *Profiler::pSingleton = NULL;

// ProfilerLocator is used for storing source code context,
// objects of this class should be static
class ProfilerLocator
{
  private:
    string Name;
    int Index;
    Profiler *pProfiler;

  public:
    // construct an object for given profiler and register new code fragment with given name (Place)
    ProfilerLocator(Profiler *ptr, string Place);
    // construct an object for default profiler (from Profiler::GetSingleton)
    // and register new code fragment with given name (Place)
    ProfilerLocator(string Place);
    
    ProfilerLocator(void);
    ~ProfilerLocator();
    
    // return descriptive name of code fragment bound to this locator
    string GetName(void) const;
    // return identifier generated for code fragment bound to this locator
    int GetIndex(void) const;
    // return host profiler object passed during construction of this locator
    Profiler *GetProfiler(void) const;
};

// ProfilerObject allows for automatic profiling starts/stops by
// instantiating and deletion automatic (on stack) objects
class ProfilerObject
{
  private:
    int id;
    Profiler *pProfiler;

  public:
    // construct new object for code fragment with id index,
    // and start profiling the code fragment in given profiler
    ProfilerObject(Profiler *ptr, int index = 0);
    // construct new object for code fragment, identified by the locator,
    // and start profiling the code fragment in the profiler bound to the locator
    ProfilerObject(const ProfilerLocator *locator);
    // construct new object for code fragment, identified by the locator,
    // and start profiling the code fragment in the profiler bound to the locator
    ProfilerObject(const ProfilerLocator &locator);
    // destruct object and stop profiling code fragment
    ~ProfilerObject();
};

//+------------------------------------------------------------------+
//| Automatic object instantiation                                   |
//+------------------------------------------------------------------+

// default profiler instance with 100 measurement points
Profiler profiler();

//+------------------------------------------------------------------+
//| Class ProfilerObject definition                                  |
//+------------------------------------------------------------------+

ProfilerObject::ProfilerObject(Profiler *ptr, int index = 0)
{
  if(CheckPointer(ptr) == POINTER_INVALID) return;
  pProfiler = ptr;
  id = index;
  pProfiler.StartProfiler(id);
}

ProfilerObject::ProfilerObject(const ProfilerLocator *locator)
{
  if(CheckPointer(locator) == POINTER_INVALID) return;
  pProfiler = locator.GetProfiler();
  id = locator.GetIndex();
  pProfiler.StartProfiler(id);
}

ProfilerObject::ProfilerObject(const ProfilerLocator &locator)
{
  pProfiler = locator.GetProfiler();
  if(CheckPointer(pProfiler) == POINTER_INVALID)
  {
    Print("ProfilerObject pProfiler POINTER_INVALID");
    return;
  }
  id = locator.GetIndex();
  pProfiler.StartProfiler(id);
}

ProfilerObject::~ProfilerObject()
{
  if(CheckPointer(pProfiler) != POINTER_INVALID)
  {
    pProfiler.StopProfiler(id);
  }
}

//+------------------------------------------------------------------+
//| Class Profiler definition                                        |
//+------------------------------------------------------------------+

Profiler::Profiler(int max = 100)
{
  ArrayResize(Start, max);
  ArrayInitialize(Start, 0);
  ArrayResize(Stop, max);
  ArrayResize(CounterTotal, max);
  ArrayInitialize(CounterTotal, 0);
  ArrayResize(DurationTotal, max);
  ArrayInitialize(DurationTotal, 0);
  ArrayResize(Names, max);
  ArrayResize(Used, max);
  ArrayInitialize(Used, false);
  MaxID = max;
  CurID = 0;
  InstanceCreateTime = GetTickCount();
  if(pSingleton == NULL) pSingleton = GetPointer(this);
  Print(__FUNCSIG__ + " " + (string)max);
}

Profiler::~Profiler()
{
  Print(__FUNCSIG__);
}

bool Profiler::StartProfiler(int id = 0)
{
  if(id < 0 || id >= MaxID) return(false);
  if(Start[id] == 0)
  {
    Start[id] = GetTickCount();
    Stop[id] = 0;
    Used[id] = true;
  }
  return(true);
}

bool Profiler::StartProfiler(const ProfilerLocator &locator)
{
  return(StartProfiler(locator.GetIndex()));
}

int Profiler::StartProfiler(string name)
{
  int id = -1;
  for(int i = 0; i < CurID; i++)
  {
    if(StringCompare(Names[i], name) == 0)
    {
      id = i;
      break;
    }
  }
  if(id == -1)
  {
    id = SetName(name);
  }
  if(Start[id] == 0)
  {
    Start[id] = GetTickCount();
    Stop[id] = 0;
    Used[id] = true;
  }
  return(id);
}

void Profiler::StopProfiling(void)
{
  for(int i = 0; i < MaxID; i++)
  {
    StopProfiler(i);
  }
}

ulong Profiler::StopProfiler(int id = 0)
{
  if(id < 0 || id >= MaxID) return(0);
  if(Start[id] == 0) return(0); // already stopped

  Used[id] = true;
  Stop[id] = GetTickCount();
  ulong delta = Stop[id] - Start[id];
  
  CounterTotal[id]++;
  DurationTotal[id] += delta;

  if(UseGlobalVariables)
  {
    string nameTime = ProfilerPrefix + GetName(id) + "-time";
    string nameCount = ProfilerPrefix + GetName(id) + "-count";
    double time = 0;
    int count = 0;
    if(GlobalVariableCheck(nameTime))
    {
      time = GlobalVariableGet(nameTime);
    }
    time += (double)delta;
    GlobalVariableSet(nameTime, time);
    
    if(GlobalVariableCheck(nameCount))
    {
      count = (int)GlobalVariableGet(nameCount);
    }
    count++;
    GlobalVariableSet(nameCount, count);
  }

  Start[id] = 0;
  Stop[id] = 0;
  
  return(delta);
}

ulong Profiler::StopProfiler(const ProfilerLocator &locator)
{
  return(StopProfiler(locator.GetIndex()));
}

ulong Profiler::GetProfilerCount(int id = 0) const
{
  if(id < 0 || id >= MaxID) return(0);
  return(CounterTotal[id]);
}

ulong Profiler::GetProfilerDuration(int id = 0) const
{
  if(id < 0 || id >= MaxID) return(0);
  return(DurationTotal[id]/1000);
}

bool Profiler::SetName(int id, string name)
{
  if(id < 0 || id >= MaxID) return(false);
  Halt(id < CurID, "Attempt to redefine profiling point " + IntegerToString(id));
  Print(__FUNCSIG__, " ", name, " ", id);
  Names[id] = name;
  Used[id] = true;
  CurID = id + 1;
  return(true);
}

int Profiler::SetName(string name)
{
  Halt(CurID >= MaxID, "Wrong index:" + IntegerToString(CurID) + " of " + IntegerToString(MaxID));
  Print(__FUNCSIG__, " ", name, " ", CurID);
  Names[CurID] = name;
  Used[CurID] = true;
  CurID++;
  return(CurID - 1);
}

string Profiler::GetName(int id) const
{
  if(id < 0 || id >= MaxID) return("?");
  return(Names[id]);
}

string Profiler::GetStats(int id = 0) const
{
  if(id < 0 || id >= MaxID) return("?");
  
  string state = "";
  ulong running = 0;
  ulong now = GetTickCount();
  if(Start[id] != 0 && Stop[id] == 0)
  {
    running = now - Start[id];
    state = "WARNING! Keeps running";
  }
  
  string result = StringFormat("%20s %5d %7d %12d %6.2f %s",
    Names[id],
    CounterTotal[id] + (running > 0),
    (DurationTotal[id] + running)/1000,
    DurationTotal[id] + running,
    (DurationTotal[id] + running) * 100.0 / (now - InstanceCreateTime),
    state
    );
  return result;
}

void Profiler::PrintAllStats(void) const
{
  Print("Profiler stats: name count seconds milliseconds     %% Comments");
  for(int i = 0; i < MaxID; i++)
  {
    if(Used[i])
    {
      Print(GetStats(i));
    }
  }
}

void Profiler::SetGlobalVariablesUse(bool use)
{
  UseGlobalVariables = use;
}


//+------------------------------------------------------------------+
//| Class ProfilerLocator definition                                 |
//+------------------------------------------------------------------+

ProfilerLocator::ProfilerLocator(Profiler *ptr, string Place)
{
  if(CheckPointer(ptr) == POINTER_INVALID)
  {
    Print(__FUNCSIG__, " POINTER_INVALID");
    return;
  }
  pProfiler = ptr;
  Name = Place;
  Index = pProfiler.SetName(Name);
  Print(__FUNCSIG__, " " + (string)pProfiler.GetCurrentID() + " " + (string)Index);
}

ProfilerLocator::ProfilerLocator(string Place)
{
  pProfiler = Profiler::GetSingleton();
  if(CheckPointer(pProfiler) == POINTER_INVALID)
  {
    Print(__FUNCSIG__, " POINTER_INVALID");
    return;
  }
  Name = Place;
  Index = pProfiler.SetName(Name);
  Print(__FUNCSIG__, " " + (string)pProfiler.GetCurrentID() + " " + (string)Index);
}

ProfilerLocator::ProfilerLocator(void)
{
  pProfiler = Profiler::GetSingleton();
  if(CheckPointer(pProfiler) == POINTER_INVALID)
  {
    Print(__FUNCSIG__, " POINTER_INVALID");
    return;
  }
  Halt(pProfiler.GetCurrentID() > 0, "Script profiler must be created once!");
  Name = MQLInfoString(MQL_PROGRAM_NAME);
  Index = pProfiler.SetName(Name);
  Print(__FUNCSIG__, " " + (string)pProfiler.GetCurrentID() + " " + (string)Index);
}

ProfilerLocator::~ProfilerLocator()
{
  Print(__FUNCSIG__);
}

string ProfilerLocator::GetName(void) const
{
  return(Name);
}

int ProfilerLocator::GetIndex(void) const
{
  return(Index);
}

Profiler *ProfilerLocator::GetProfiler(void) const
{
  return(pProfiler);
}

//+------------------------------------------------------------------+
//| Assert                                                           |
//+------------------------------------------------------------------+

void Halt(bool Condition, string Message)
{
  int x = 0;
  if(Condition)
  {
    Print(Message);
    x++;
    x--;
    x = 1 / x;
  }
}

// USAGE
/*

#include <profiler.mqh>

#define CALC_PROFILE_ID 10

_PROFILESCRIPT_;

void subfunc()
{
  _PROFILEFUNC_;
  
  for(int i = 0; i < 1000; i++)
  {
    profiler.StartProfiler(CALC_PROFILE_ID);
    for(int j = 0; j < 10000; j++)
    {
      MathSqrt(MathPow(i, 3));
    }
    profiler.StopProfiler(CALC_PROFILE_ID);
  }

  Sleep(250);
}

void OnStart()
{
  profiler.SetName(CALC_PROFILE_ID, "calculation");
  Sleep(500);
  subfunc();
  profiler.PrintAllStats();
}


// OUTPUT EXAMPLE

Profiler stats: name count seconds milliseconds     %% Comments
        profilertest     1       1         1607 100.00 WARNING! Keeps running
             subfunc     1       1         1107  68.89 
         calculation  1000       0          858  53.39 

*/